/*
 * libbsr - allocation services for the Barrier Synchronization Register
 *          on IBM POWER systems
 *
 * Copyright (C) 2008 Sonny Rao, IBM Corporation.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "bsr.h"

static char *bsr_lockdir  = "/var/lock/";
static char *bsr_sysfsdir = "/sys/class/bsr/";


struct bsr_info {
	unsigned num_devs;
	unsigned max_size;
	unsigned num_sizes;
	unsigned sizemap;
	unsigned free;
	unsigned freemap;
/* enumerate all power of 2 sizes up to 8 MB */
/* so we can keep a count of how many there are */
#define MAX_BSR_SIZE 23
/* support up to 256 (8 bits) of each type*/
#define MAX_CLASS_DEVS 256
	unsigned counts[MAX_BSR_SIZE + 1];
};


/* we'll cache this info */
static struct bsr_info bsr_all;


/* this tells us that there are two BSR sizes */
/* one is 8 bytes and the other is larger and */
/* the 8 byte BSRs map ontop of the larger one */
/* so if the big one is taken, the smaller ones */
/* are also taken */
/* this is the source of much ugliness in the code */
static unsigned bsr_alias_mode;
static unsigned bsr_big_size;

/* our idea of "total bytes" */
static unsigned bsr_total_bytes;

static int bsr_log2(unsigned num)
{
	int ret = 0;

	if (!num)
		return ret;

	while (num) {
		ret++;
		num = num >> 1;
	}
	return ret - 1;
}


static int bsr_query_dir(char *dirname, struct bsr_info *bsr_info)
{
	struct dirent *ent;
	DIR *sysdir;

	sysdir = opendir(dirname);	
	if (!sysdir)
		return ENODEV;
	memset(bsr_info, 0, sizeof(*bsr_info));
	
	while ((ent = readdir(sysdir)) != NULL) {
		unsigned size, id, class;			
		int count = sscanf(ent->d_name,"bsr%d_%d",&size, &id);
		if (count == 2) {
			bsr_info->num_devs++;
			if (bsr_info->max_size < size)
				bsr_info->max_size = size;
			
				class = bsr_log2(size);
				bsr_info->counts[class]++;
				
				if (!(bsr_info->sizemap & (1 << class))) {
					bsr_info->num_sizes++;
					bsr_info->sizemap |= (1 << class);
				}
		}

	}
	closedir(sysdir);
	return 0;
}

static unsigned total_counts(struct bsr_info *bsi)
{
	int i;
	unsigned total = 0;
	for (i=0 ; i < MAX_BSR_SIZE; i++) {
		unsigned bytes = 1 << i;
		unsigned all = bsi->counts[i] * bytes;
		total += all;
	}
	return total;
}

int bsr_query (unsigned *total_bytes, uint32_t *supp_window_mask, 
	       unsigned *free_bytes, uint32_t *free_window_mask)
{
	int ret;
	struct bsr_info bsr_locks;

	/* only read the total info once */
	if (!bsr_all.num_devs) {
		ret = bsr_query_dir(bsr_sysfsdir, &bsr_all);
		if (ret)
			return ret;

		/* if there are two sizes and one is 8bytes*/
		/* the smaller ones map over the big one */
		if ((bsr_all.num_sizes == 2) &&
		    (bsr_all.sizemap & (1 << 3))) {
			bsr_alias_mode = 1;
			bsr_big_size  = bsr_all.sizemap & (~(1 << 3));
			bsr_total_bytes = bsr_big_size;
		} else {
			bsr_total_bytes = total_counts(&bsr_all);
		}
	}

	*supp_window_mask = bsr_all.sizemap;
	*total_bytes = bsr_total_bytes;

	/* look at the lockdir and figure out how many are free */
	ret = bsr_query_dir(bsr_lockdir, &bsr_locks);
	if (ret)
		return ret;

	/* alias mode is messy.. if the big one is taken, they're */
	/* all taken, also if any small one is taken, then the big one is */
	/* also taken */
	if (bsr_alias_mode) {
		if (bsr_locks.sizemap & bsr_big_size) {
			*free_bytes = 0;
			*free_window_mask = 0;
		} else if (bsr_locks.sizemap & 8) {
			/* someone has a small one */
			*free_bytes = bsr_big_size - 
				total_counts(&bsr_locks);
			*free_window_mask = 8;			
		} else {
			/* everything is free */ 
			*free_bytes = bsr_total_bytes;
			*free_window_mask = bsr_all.sizemap;
		}
	} else {
		int i;
		unsigned free_mask  = 0;
		*free_bytes = bsr_total_bytes - 
			total_counts(&bsr_locks);
		for ( i=0; i < MAX_BSR_SIZE; i++) {
			if (bsr_all.counts[i]) {
				if (bsr_locks.counts[i] < bsr_all.counts[i])
					free_mask = 1 << i;
			}
		}
		*free_window_mask = free_mask;
	}	
	return 0;	
}

static int bsr_alloc_size(unsigned size, unsigned *id)
{
	struct dirent *ent;
	DIR *dir;
	unsigned class, class_total, i, found = 0;
	char tmpl[32];
	char alloced[MAX_CLASS_DEVS];

	class = bsr_log2(size);
	class_total = bsr_all.counts[class];

	memset(alloced, 0, sizeof(alloced));

	snprintf(tmpl, 32, "bsr%d_%%d",size);

	dir = opendir(bsr_lockdir);
	while ((ent = readdir(dir)) != NULL) {
		unsigned num;
		if (ent->d_type == DT_REG) {
			if (sscanf(ent->d_name, tmpl, &num) == 1){
				alloced[num] = 0xff;
				found++;
			}
		}
	}
	closedir(dir);
/* 	printf("found %d in-use\n",found); */
	for (i=0 ; i < class_total;i++) {
		if (!alloced[i]) {
			char buf[32];
			char pidbuf[16];
			int fd;
			pid_t pid;
			snprintf(buf, 32,"%sbsr%d_%d",bsr_lockdir,size, i);
			fd = open(buf, O_CREAT | O_EXCL | O_RDWR,  S_IRWXU);
			if (fd < 0) 
				continue;
			pid = getpid();
			snprintf(pidbuf, 16,"%10d\n",pid);
			if (write(fd, pidbuf,11) < 11) {
				unlink(buf);
				close(fd);
				return EIO;
			}
			close(fd);
			*id = i;
			return 0;
		}
	}
	return EBUSY;	
}

static int bsr_get_attr(unsigned size, unsigned id, char *name,
			unsigned *bsr_attr)
{
	char buf[64];
	int fd;
	snprintf(buf,64, "%sbsr%d_%d/%s", bsr_sysfsdir,size,id, name);
	if ((fd = open(buf, O_RDONLY)) < 0)
		return errno;
	if (read(fd, buf, 64) < 0){
		close(fd);
		return EIO;
	}
	close(fd);
	sscanf(buf,"%d", bsr_attr);
	return 0;
}

int bsr_alloc (unsigned bsr_bytes, unsigned *bsr_stride, 
	       unsigned *bsr_id)
{
	unsigned total, allmask, free, freemask, id;
	int ret = bsr_query(&total, &allmask, &free, &freemask);
	unsigned class = bsr_log2(bsr_bytes);

	if (ret)
		return ret;

	if (!free || !(freemask & bsr_bytes))
		return EBUSY;

	ret = bsr_alloc_size(bsr_bytes, &id);
	if (ret)
		return ret;

	ret = bsr_get_attr(bsr_bytes, id, "bsr_stride", bsr_stride);
	if (ret) {
		bsr_free(id);
		return ret;
	}
	*bsr_id = ((class << 24) | id);
	return 0;
}

/* TODO, more checking ? */
int bsr_free(unsigned bsr_id)
{
	char buf[64];
	unsigned class = (bsr_id >> 24);
	unsigned id = bsr_id & 0xff;
	unsigned bytes = 1 << class ;
	int ret;

	snprintf(buf, 64, "%sbsr%d_%d", bsr_lockdir, bytes, id);
	ret = unlink(buf);
		
	return ret;
}

void *bsr_map(void *addr, unsigned bsr_id, int prot, int flags, unsigned *bsr_length)
{
	char buf[64];
	unsigned class = (bsr_id >> 24);
	unsigned bytes = 1 << class;
	unsigned id = bsr_id & 0xff;
	unsigned length = 0;
	int fd, ret;
	void * ret_addr;

	ret = bsr_get_attr(bytes, id, "bsr_length", &length);
	if (ret)
		return NULL;

	*bsr_length = length;

	snprintf(buf, 64, "%sbsr%d_%d", "/dev/", bytes,id);
	fd = open(buf, O_RDWR);
	if (fd < 0)
		return NULL;

	if (!prot) 
		prot = PROT_READ | PROT_WRITE;
	if (!flags)
		flags = MAP_SHARED;

	ret_addr = mmap(addr, length, prot, flags, fd, 0);
	close(fd);
	return ret_addr;
}

